@@ -538,29 +538,23 @@ module Parser
     end
 
     def associate(begin_t, pairs, end_t)
-      0.upto(pairs.length - 1) do |i|
-        (i + 1).upto(pairs.length - 1) do |j|
-          key1, = *pairs[i]
-          key2, = *pairs[j]
+      key_set = Set.new
 
-          do_warn = false
+      pairs.each do |pair|
+        next unless pair.type.eql?(:pair)
 
-          # keys have to be simple nodes, MRI ignores equal composite keys like
-          # `{ a(1) => 1, a(1) => 1 }`
-          case key1.type
-          when :sym, :str, :int, :float
-            if key1 == key2
-              do_warn = true
-            end
-          when :rational, :complex, :regexp
-            if @parser.version >= 31 && key1 == key2
-              do_warn = true
-            end
-          end
+        key, = *pair
 
-          if do_warn
-            diagnostic :warning, :duplicate_hash_key, nil, key2.loc.expression
-          end
+        case key.type
+        when :sym, :str, :int, :float
+        when :rational, :complex, :regexp
+          next unless @parser.version >= 31
+        else
+          next
+        end
+
+        unless key_set.add?(key)
+          diagnostic :warning, :duplicate_hash_key, nil, key.loc.expression
         end
       end
 
@@ -600,7 +594,13 @@ module Parser
     end
 
     def gvar(token)
-      n(:gvar, [ value(token).to_sym ],
+      gvar_name = value(token)
+
+      if gvar_name.start_with?('$0') && gvar_name.length > 2
+        diagnostic :error, :gvar_name, { :name => gvar_name }, loc(token)
+      end
+
+      n(:gvar, [ gvar_name.to_sym ],
         variable_map(token))
     end
 
@@ -660,6 +660,13 @@ module Parser
         end
 
         unless @parser.static_env.declared?(name)
+          if @parser.version == 33 &&
+              name == :it &&
+              @parser.context.in_block &&
+              !@parser.max_numparam_stack.has_ordinary_params?
+            diagnostic :warning, :ambiguous_it_call, nil, node.loc.expression
+          end
+
           return n(:send, [ nil, name ],
             var_send_map(node))
         end
@@ -1696,7 +1703,19 @@ module Parser
           cond
         end
 
-      when :and, :or, :irange, :erange
+      when :and, :or
+        lhs, rhs = *cond
+
+        if @parser.version == 18
+          cond
+        else
+          cond.updated(cond.type, [
+            check_condition(lhs),
+            check_condition(rhs)
+          ])
+        end
+
+      when :irange, :erange
         lhs, rhs = *cond
 
         type = case cond.type
@@ -1704,15 +1723,13 @@ module Parser
         when :erange then :eflipflop
         end
 
-        if [:and, :or].include?(cond.type) &&
-               @parser.version == 18
-          cond
-        else
-          cond.updated(type, [
-            check_condition(lhs),
-            check_condition(rhs)
-          ])
-        end
+        lhs_condition = check_condition(lhs) unless lhs.nil?
+        rhs_condition = check_condition(rhs) unless rhs.nil?
+
+        return cond.updated(type, [
+          lhs_condition,
+          rhs_condition
+        ])
 
       when :regexp
         n(:match_current_line, [ cond ], expr_map(cond.loc.expression))
@@ -2251,6 +2268,10 @@ module Parser
 
     def static_regexp_node(node)
       if node.type == :regexp
+        if @parser.version >= 33 && node.children[0..-2].any? { |child| child.type != :str }
+          return nil
+        end
+
         parts, options = node.children[0..-2], node.children[-1]
         static_regexp(parts, options)
       end
